from os import linesep as nl


def terminal_whitespace(inputfile, from_file=True):
    """
    Return the number of whitespace newlines in the file.

    If the file contains no newlines, returns 0.
    """
    if from_file:
        with open(inputfile, "r") as f:
            lines = f.readlines()
    else:
        lines = inputfile
    terminal_i = -1
    for i, line in enumerate(reversed(lines)):
        if line.rstrip(nl) == "":
            if i - terminal_i > 1:
                break
            else:
                # Accumulate consecutive whitespace index until returning
                terminal_i = i
    return terminal_i + 1



import numpy as np
from numpy import arange, pi
import matplotlib.pyplot as plt
from os.path import (
    basename as aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa,
    sep as pathsep,
    islink,
)


def show_line(n=4, suppress_display=False):
    assert type(n) in (int, np.typeDict["int"]), f"n must be an integer, {n} is not"
    assert n > 0, "Please provide a positive integer, {n} is not"
    l = arange(0, n + 1)
    if suppress_display:
        return plt.plot(l)
    plt.plot(l)
    plt.show()
    return


def print_some_pi(n=2, suppress_print=False):
    assert type(n) in (int, np.typeDict["int"]), f"n must be an integer, {n} is not"
    message = f"hey{pathsep}hi"
    output = f"{message}, {n} pi = {n * pi}"
    if suppress_print:
        return output
    print(output)
    return



import numpy as np
from example.demo_program import show_line, print_some_pi
from os.path import sep


def test_show_line(n=4):
    plot_output = show_line(n, suppress_display=True)
    assert type(plot_output) is list and len(plot_output) == 1
    l = plot_output[0]
    plot_path = l.get_path()
    ext = plot_path.get_extents()
    assert np.array_equal(ext.p0, np.array([0.0, 0.0]))
    assert np.array_equal(ext.p1, np.array([float(n), float(n)]))
    return True


def test_print_some_pi(n=2):
    output = print_some_pi(n, suppress_print=True)
    assert output == sep.join(["hey", "hi"]) + ", 2 pi = 6.283185307179586"
    return True


def get_test_failures():
    exceptions = []
    try:
        assert test_show_line(), "Test failed: show_line"
    except AssertionError as e:
        exceptions.append(e)
    try:
        assert test_print_some_pi(), "Test failed: print_some_pi"
    except AssertionError as e:
        exceptions.append(e)
    if exceptions is []:
        return None
    else:
        return exceptions


def list_failing_tests():
    """
    Returns None if no failing tests, otherwise returns a list of
    the names of the functions whose tests failed (N.B. not the name
    of the functions doing the testing, but the ones they test).
    """
    exceptions = get_test_failures()
    if exceptions == []:
        return
    else:
        failed_funcs = []
        for e in exceptions:
            if e.startswith("Test failed:"):
                fail_func = e.split(":")[1][1:]
                failed_funcs.append(fail_func)
        return failed_funcs


def test_report(verbose=True):
    """
    Give an error if any tests are failing, otherwise return None.
    """
    failing = list_failing_tests()
    assert failing is None, f"Tests failed for {failing}"
    if verbose:
        print("✔ All tests pass")
    return



from pathlib import Path
from example import __path__ as example_dir
from mvdef import __path__ as module_dir, __package__ as module_name

try:
    assert len(module_dir) == 1
    module_dir = Path(module_dir[0])
except AssertionError as e:
    raise NotImplementedError(e, "There should only be one module directory here")

try:
    assert module_dir.exists() and module_dir.is_dir()
except AssertionError as e:
    raise ValueError(e, "Module directory should be a directory that already exists")

try:
    assert module_dir.name == module_name
except AssertionError as e:
    raise ValueError(e, "Module directory name doesn't match record of module name")

try:
    assert len(example_dir) == 1
    example_dir = module_dir.parent / example_dir[0]
except AssertionError as e:
    raise NotImplementedError(e, "There should only be one example path here")

try:
    assert example_dir.exists() and example_dir.is_dir()
except AssertionError as e:
    raise ValueError(e, "Example directory should be a directory that already exists")

# Exports: (note: the module_dir and src_dir might need switching after renaming?)
#   - module_dir: absolute Path to the module directory, `src/`,
#                 containing `__main__.py`, `mvdef/`, and `example/`
#   - module_name: string giving the module name (`"mvdef"`)
#   - example_dir: absolute Path to the `example/` directory



from mvdef.colours import colour_str as colour


def pprint_agenda_desc(category, entry_key, entry_dict, extra_message=""):
    """
    Pretty print an edit agenda entry according to the agenda category.
    The 7 categories are: move, keep, copy, lose, take, echo, stay.
    """
    entry_desc = describe_def_name_dict(entry_key, entry_dict)
    if category == "move":
        m = colour("green", f" ⇢ MOVE  ⇢ {entry_desc}" + extra_message)
    elif category == "keep":
        m = colour("dark_gray", f"⇠  KEEP ⇠  {entry_desc}" + extra_message)
    elif category == "copy":
        m = colour("light_blue", f"⇠⇢ COPY ⇠⇢ {entry_desc}" + extra_message)
    elif category == "lose":
        m = colour("red", f" ✘ LOSE ✘  {entry_desc}" + extra_message)
    elif category == "take":
        m = colour("green", f" ⇢ TAKE  ⇢ {entry_desc}" + extra_message)
    elif category == "echo":
        m = colour("light_blue", f"⇠⇢ ECHO ⇠⇢ {entry_desc}" + extra_message)
    elif category == "stay":
        m = colour("dark_gray", f"⇠  STAY ⇠  {entry_desc}" + extra_message)
    else:
        raise ValueError(f"Unknown agenda category: {category}")
    print(m)
    return


def pprint_agenda(agenda):
    for category in agenda:
        for entry in agenda.get(category):
            name, info_dict = list(entry.items())[0]
            pprint_agenda_desc(category, name, info_dict)
    return


def describe_def_name_dict(name, name_dict):
    """
    Wrapper function that returns a string presenting the content of a dict entry
    with import statement indexes, line number, and import source path. These
    fields are instantiated within `get_def_names`, which in turn is assigned to
    the variable `mvdef_names` within `parse_mv_funcs`.
    
    The output of `parse_mv_funcs` gets passed to `process_ast`, which iterates over
    the subsets within the output of `parse_mv_funcs`, at which stage it's
    necessary to produce a nice readable output, calling `describe_mvdef_name_dict`.
    """
    # Extract: import index; intra-import index; line number; import source
    n, n_i, ln, imp_src = [name_dict.get(x) for x in ["n", "n_i", "line", "import"]]
    desc = f"(import {n}:{n_i} on line {ln}) {name} ⇒ <{imp_src}>"
    return desc



import ast
from asttokens import ASTTokens
import builtins
from pathlib import Path
from mvdef.agenda_util import pprint_agenda
from mvdef.deprecations import pprint_def_names
from mvdef.import_util import get_imported_name_sources, annotate_imports, imp_def_subsets


def ast_parse(fp, mvdefs=[], transfers={}, report=True):
    """
    Build and arse the Abstract Syntax Tree (AST) of a Python file, and either return
    a report of what changes would be required to move the mvdefs subset of all
    function definitions out of it, or a report of the imports and funcdefs in general
    if no mvdefs is provided (taken to indicate that the file is the target funcdefs
    are moving to), or make changes to the file (either newly creating one if no such
    file exists, or editing in place according to the reported import statement
    differences).

    If the Python file `fp` doesn't exist (which can be checked directly as it's a
    Path object), it's being newly created by the move and obviously no report can
    be made on it: it has no funcdefs and no import statements, so all the ones being
    moved will be newly created.

    mvdefs should be given if the file is the source of moved functions, and left
    empty (defaulting to value of []) if the file is the destination to move them to.
    
    If report is True, returns a string describing the changes
    to be made (if False, nothing is returned).
    
    If backup is True, files will be changed in place by calling mvdef.backup.backup
    (obviously, be careful switching this setting off if report is True, as any
    changes made cannot be restored afterwards from this backup file).
    """
    extant = fp.exists() and fp.is_file()
    if extant:
        with open(fp, "r") as f:
            fc = f.read()
            # a = ast
            trunk = ast.parse(fc).body

        # return imports, funcdefs
        edit_agenda = process_ast(fp, mvdefs, trunk, transfers, report)
        return edit_agenda
    elif mvdefs == []:
        # not extant so file doesn't exist (cannot produce a parsed AST)
        # however mvdefs is [] so file must be dst, return value of None
        return
    else:
        raise ValueError(f"Can't move {mvdefs} from {fp} – it doesn't exist!")
    return


def process_ast(fp, mvdefs, trunk, transfers={}, report=True):
    """
    Handle the hand-off to dedicated functions to go from the mvdefs of functions
    to move, first deriving lists of imported names which belong to the mvdefs and
    the non-mvdefs functions (using `parse_mv_funcs`), then constructing an
    'edit agenda' (using `process_ast`) which describes [and optionally
    reports] the changes to be made at the file level, in terms of move/keep/copy
    operations on individual import statements between the source and destination
    Python files.

      mvdefs:     List of functions to be moved
      trunk:      Tree body of the file's AST, which will be separated into
                  function definitions, import statements, and anything else.
      transfers:  List of transfers already determined to be made from the src
                  to the dst file (from the first call to ast_parse)
      report:     Whether to print a report during the program (default: True)

    -------------------------------------------------------------------------------
    
    First, given the lists of mvdef names (m_names) and non-mvdef names
    (nm_names), construct the subsets:

      mv_imps:  imported names used by the functions to move (only in mvdef_names),
      nm_imps:  imported names used by the functions not to move (only in
                nonmvdef_names),
      mu_imps:  imported names used by both the functions to move and the
                functions not to move (in both mvdef_names and nonmvdef_names)

    Potentially 'as a dry run' (if this is being called by process_ast and its
    parameter edit is False), report how to remove the import statements or statement
    sections which import mv_inames, do nothing to the import statements which import
    nonmv_inames, and copy the import statements which import mutual_inames (as both
    src and dst need them).

    Additionally, accept 'transfers' from a previously determined edit agenda,
    so as to "take" the "move" names, and "echo" the "copy" names (i.e. when
    receiving names marked by "move" and "copy", distinguish them to indicate
    they are being received by transfer [from src⇒dst file], for clarity).

    For clarity, note that this function does **not** edit anything itself, it just
    describes how it would be possible to carry out the required edits at the level
    of Python file changes.
    """
    # get_edit_agenda(m_names, nm_names, rm_names, transfers, report=True)
    m_names, nm_names, rm_names = parse_mv_funcs(mvdefs, trunk, report=report)
    imported_names = get_imported_name_sources(trunk, report=report)
    if report:
        print(f"• Determining edit agenda for {fp.name}:")
    agenda_categories = ["move", "keep", "copy", "lose", "take", "echo", "stay"]
    agenda = dict([(c, []) for c in agenda_categories])
    # mv_inames is mv_imports returned from imp_def_subsets, and so on
    mv_imps, nm_imps, mu_imps = imp_def_subsets(m_names, nm_names, report=report)
    # Iterate over each imported name, i, in the subset of import names to move
    for i in mv_imps:
        assert i in set().union(*[m_names.get(k) for k in m_names]), f"{i} not found"
        i_dict = [m_names.get(k) for k in m_names if i in m_names.get(k)][0].get(i)
        agenda.get("move").append({i: i_dict})
    for i in nm_imps:
        assert i in set().union(*[nm_names.get(k) for k in nm_names]), f"{i} not found"
        i_dict = [nm_names.get(k) for k in nm_names if i in nm_names.get(k)][0].get(i)
        agenda.get("keep").append({i: i_dict})
    for i in mu_imps:
        assert i in set().union(*[m_names.get(k) for k in m_names]), f"{i} not found"
        i_dict = [m_names.get(k) for k in m_names if i in m_names.get(k)][0].get(i)
        agenda.get("copy").append({i: i_dict})
    for i in rm_names:
        i_dict = rm_names.get(i)
        agenda.get("lose").append({i: i_dict})
    if transfers == {}:
        # Returning without transfers
        if report:
            pprint_agenda(agenda)
        return agenda
    # elif report:
    #    if len(agenda.get("lose")) > 0:
    #        print("• Resolving edit agenda conflicts:")
    # i is 'ready made' from a previous call to ast_parse, and just needs reporting
    for i in transfers.get("take"):
        k, i_dict = list(i.items())[0]
        agenda.get("take").append({k: i_dict})
    for i in transfers.get("echo"):
        k, i_dict = list(i.items())[0]
        agenda.get("echo").append({k: i_dict})
    # Resolve agenda conflicts: if any imports marked 'lose' are cancelled out
    # by any identically named imports marked 'take' or 'echo', change to 'stay'
    for i in agenda.get("lose"):
        k, i_dict = list(i.items())[0]
        imp_src = i_dict.get("import")
        if k in [list(x)[0] for x in agenda.get("take")]:
            t_i_dict = [x for x in agenda.get("take") if k in x][0].get(k)
            take_imp_src = t_i_dict.get("import")
            if imp_src != take_imp_src:
                continue
            # Deduplicate 'lose'/'take' k: replace both with 'stay'
            agenda.get("stay").append({k: i_dict})
            l_k_i = [list(x.values())[0] for x in agenda.get("lose")].index(i_dict)
            del agenda.get("lose")[l_k_i]
            t_k_i = [list(x.values())[0] for x in agenda.get("take")].index(t_i_dict)
            del agenda.get("take")[t_k_i]
        elif k in [list(x)[0] for x in agenda.get("echo")]:
            e_i_dict = [x for x in agenda.get("echo") if k in x][0].get(k)
            echo_imp_src = e_i_dict.get("import")
            if imp_src != echo_imp_src:
                continue
            # Deduplicate 'lose'/'echo' k: replace both with 'stay'
            agenda.get("stay").append({k: i_dict})
            l_k_i = [list(x.values())[0] for x in agenda.get("lose")].index(i_dict)
            del agenda.get("lose")[l_k_i]
            e_k_i = [list(x.values())[0] for x in agenda.get("echo")].index(e_i_dict)
            del agenda.get("echo")[e_k_i]
    # Resolve agenda conflicts: if any imports marked 'take' or 'echo' are cancelled
    # out by any identically named imports already present, change to 'stay'
    for i in agenda.get("take"):
        k, i_dict = list(i.items())[0]
        take_imp_src = i_dict.get("import")
        if k in imported_names:
            # Check the import source and asnames match
            imp_src = imported_names.get(k)[0]
            if imp_src != take_imp_src:
                # This means that the same name is being used by a different function
                raise ValueError(
                    f"Cannot move imported name '{k}', it is already "
                    + f"in use in {fp.name} ({take_imp_src} clashes with {imp_src})"
                )
                # (N.B. could rename automatically as future feature)
            # Otherwise there is simply a duplicate import statement, so the demand
            # to 'take' the imported name is already fulfilled.
            # Replace unnecessary 'take' with 'stay'
            agenda.get("stay").append({k: i_dict})
            t_k_i = [list(x.values())[0] for x in agenda.get("take")].index(i_dict)
            del agenda.get("take")[t_k_i]
    for i in agenda.get("echo"):
        k, i_dict = list(i.items())[0]
        echo_imp_src = i_dict.get("import")
        if k in imported_names:
            # Check the import source and asnames match
            imp_src = imported_names.get(k)[0]
            if imp_src != echo_imp_src:
                # This means that the same name is being used by a different function
                raise ValueError(
                    f"Cannot move imported name '{k}', it is already "
                    + f"in use in {fp.name} ({echo_imp_src} clashes with {imp_src})"
                )
                # (N.B. could rename automatically as future feature)
            # Otherwise there is simply a duplicate import statement, so the demand
            # to 'echo' the imported name is already fulfilled.
            # Replace unnecessary 'take' with 'stay'
            agenda.get("stay").append({k: i_dict})
            e_k_i = [list(x.values())[0] for x in agenda.get("echo")].index(i_dict)
            del agenda.get("echo")[e_k_i]
    if report:
        pprint_agenda(agenda)
    return agenda


def find_assigned_args(fd):
    """
    Produce a list of the names in a function definition `fd` which are created
    by assignment operations (as identified via the function definition's AST).
    """
    args_indiv = []  # Arguments assigned individually, e.g. x = 1
    args_multi = []  # Arguments assigned from a tuple, e.g. x, y = (1,2)
    for a in ast.walk(fd):
        if type(a) is ast.Assign:
            # Handle explicit assignments from use of the equals symbol
            assert len(a.targets) == 1, "Expected 1 target per ast.Assign"
            if type(a.targets[0]) is ast.Name:
                args_indiv.append(a.targets[0].id)
            elif type(a.targets[0]) is ast.Tuple:
                args_multi.extend([x.id for x in a.targets[0].elts])
        elif type(a) is ast.For:
            # Handle implicit assignments (ctx = Store) within for loops
            if type(a.target) is ast.Name:
                args_indiv.append(a.target.id)
            elif type(a.target) is ast.Tuple:
                for x in a.target.elts:
                    assert type(x) in [ast.Name, ast.Tuple], f"Unexpected target.elts"
                    if type(x) is ast.Name:
                        args_multi.append(x.id)
                    else:  # type(x) is ast.Tuple
                        for y in x.elts:
                            assert type(y) is ast.Name, f"Unexpected target.elts tuple"
                            args_multi.append(y.id)
            else:
                raise ValueError(f"{a.target} lacks the expected ast.Name statements")
        elif type(a) is ast.ListComp:
            # Handle implicit assignments within list comprehensions
            for g in a.generators:
                assert type(g) is ast.comprehension, f"{a} not ast.comprehension type"
                if type(g.target) is ast.Name:
                    args_indiv.append(g.target.id)
                elif type(g.target) is ast.Tuple:
                    args_multi.extend([x.id for x in g.target.elts])
                else:
                    raise ValueError(f"{g.target} lacks expected ast.Name statements")
        elif type(a) is ast.Lambda:
            if len(a.args.args) > 1:
                args_multi.extend([r.arg for r in a.args.args])
            else:
                assert len(a.args.args) == 1, "A lambda can't assign no names (can it?)"
                args_indiv.append(a.args.args[0].arg)
        elif type(a) is ast.NamedExpr:
            # This is the walrus operator `:=` added in Python 3.8
            assert type(a.target) is ast.Name, f"Expected a name for {a.target}"
            args_indiv.append(a.target.id)
            # I haven't seen a multiple assignment from a walrus operator, asked SO:
            # https://stackoverflow.com/q/59567172/2668831
    assigned_args = args_indiv + args_multi
    return assigned_args


def get_extradef_names(extra_nodes):
    """
    Return the names used in the AST trunk nodes which are outside of both function
    definitions and import statements, so as to distinguish the unused names from
    those which are just used outside of function definitions.
    """
    extradef_names = set()
    for node in extra_nodes:
        node_names = [x.id for x in list(ast.walk(node)) if type(x) is ast.Name]
        for n in node_names:
            extradef_names.add(n)
    return extradef_names


def get_nondef_names(unused, import_annos, report=True):
    imp_name_lines, imp_name_dicts = import_annos
    # nondef_names is a dictionary keyed by the unused names (which were imported)
    nondef_names = dict([(x, {}) for x in unused])
    unknowns = [n for n in unused if n not in imp_name_lines]
    assert unknowns == [], f"These names could not be sourced: {unknowns}"
    # mv_imp_refs is the subset of imp_name_lines for movable funcdef names
    # These refs will lead to import statements being copied and/or moved
    uu_imp_refs = dict([[n, imp_name_lines.get(n)] for n in unused])
    for k in uu_imp_refs:
        n = uu_imp_refs.get(k).get("n")
        d = imp_name_dicts[n]
        n_i = [list(d.keys()).index(x) for x in d if d[x] == k][0]
        assert n_i >= 0, f"Movable name {k} not found in import name dict"
        # Store index in case of multiple imports per import statement line
        uu_imp_refs.get(k)["n_i"] = n_i
        uu_name_entry = nondef_names.get(k)
        n = uu_imp_refs.get(k).get("n")
        uu_name_entry["n"] = n
        uu_name_entry["n_i"] = n_i
        uu_name_entry["line"] = uu_imp_refs.get(k).get("line")
        uu_name_entry["import"] = list(imp_name_dicts[n].keys())[n_i]
    return nondef_names


def get_def_names(func_list, funcdefs, import_annos, extradef_names, report=True):
    imp_name_lines, imp_name_dicts = import_annos
    def_names = dict([(x, {}) for x in func_list])
    for m in func_list:
        fd_names = set()
        assert m in [f.name for f in funcdefs], f"No function '{m}' is defined"
        fd = funcdefs[[f.name for f in funcdefs].index(m)]
        fd_ids = [f.name for f in funcdefs]
        fd_params = [a.arg for a in fd.args.args]
        assigned_args = find_assigned_args(fd)
        for ast_statement in fd.body:
            for node in list(ast.walk(ast_statement)):
                if type(node) == ast.Name:
                    n_id = node.id
                    if n_id not in dir(builtins) + fd_ids + fd_params + assigned_args:
                        if n_id not in extradef_names:
                            fd_names.add(n_id)
        def_names[m] = dict([(x, {}) for x in sorted(fd_names)])
        # All names successfully found and can finish if remaining names are
        # in the set of funcdef names, comparing them tothe import statements
        unknowns = [n for n in fd_names if n not in imp_name_lines]
        assert unknowns == [], f"These names could not be sourced: {unknowns}"
        # mv_imp_refs is the subset of imp_name_lines for movable funcdef names
        # These refs will lead to import statements being copied and/or moved
        mv_imp_refs = dict([(n, imp_name_lines.get(n)) for n in fd_names])
        for k in mv_imp_refs:
            n = mv_imp_refs.get(k).get("n")
            d = imp_name_dicts[n]
            n_i = [list(d.keys()).index(x) for x in d if d[x] == k][0]
            assert n_i >= 0, f"Movable name {k} not found in import name dict"
            # Store index in case of multiple imports per import statement line
            mv_imp_refs.get(k)["n_i"] = n_i
            fd_name_entry = def_names.get(m).get(k)
            n = mv_imp_refs.get(k).get("n")
            fd_name_entry["n"] = n
            fd_name_entry["n_i"] = n_i
            fd_name_entry["line"] = mv_imp_refs.get(k).get("line")
            fd_name_entry["import"] = list(imp_name_dicts[n].keys())[n_i]
    return def_names


def parse_mv_funcs(mvdefs, trunk, report=True):
    """
    mvdefs:  the list of functions to move (string list of function names)
    trunk:   AST body for the file (via `ast.parse(fc).body`)
    report:  whether to print [minimal, readable] 'reporting' output

    Produce a dictionary, `mvdef_names`, whose keys are the list of functions
    to move (i.e. the list `mvdefs` becomes the list of keys of `mvdef_names`),
    and the value of which at each key (for a key `m` which indicates the name
    of one of the functions given in `mvdefs` to move) is another dictionary,
    keyed by the full set of names used in that function (`m`) which rely upon
    import statements (i.e. are not builtin names nor passed as parameters to
    the function, nor assigned in the body of the function), and the value
    of which at each name is a final nested dictionary whose keys are always:
      n:      The [0-based] index of the name's source import statement in the
              AST list of all ast.Import and ast.ImportFrom.
      n_i:    The [0-based] index of the name's source import statement within
              the one or more names imported by the source import statement at
              index n (e.g. for `pi` in `from numpy import e, pi`, `n_i` = 1).
      line:   The [1-based] line number of the import statement as given in its
              corresponding AST entry, which indicates the line number of the
              `import` call, not [necessarily] that of the name (i.e. the name
              may not be located there in the file for multi-line imports).
      import: The path of the import statement, which may contain multiple
              parts conjoined by `.` (e.g. `matplotlib.pyplot`)
    
    I.e. the dictionary with entries accessed as `mvdef_names.get(m).get(k)`
    for `m` in `mvdefs` and `k` in the subset of AST-identified imported names
    in the function with  if f.name not in mvdefs name `m` in the list of
    function definitions `defs`. This access is handed off to the helper
    function `get_def_names`.

    For the names that were imported but not used, the dictionary is not keyed
    by function (as there are no associated functions), and instead the entries
    are accessed as `nondef_names.get(k)` for `k` in `unused_names`. This access
    is handed off to the helper function `get_nondef_names`.
    """
    report_VERBOSE = False  # Silencing debug print statements
    import_types = [ast.Import, ast.ImportFrom]
    imports = [n for n in trunk if type(n) in import_types]
    defs = [n for n in trunk if type(n) == ast.FunctionDef]
    # Any nodes in the AST that aren't imports or defs are 'extra' (as in 'other')
    extra = [n for n in trunk if type(n) not in [*import_types, ast.FunctionDef]]
    # Omit names used outside of function definitions so as not to remove them
    extradefs = get_extradef_names(extra)
    if report_VERBOSE:
        print("extra:", extra)
    import_annos = annotate_imports(imports, report=report)
    mvdef_names = get_def_names(mvdefs, defs, import_annos, extradefs, report)
    if report_VERBOSE:
        print("mvdef names:")
        pprint_def_names(mvdef_names)
    # ------------------------------------------------------------------------ #
    # Next obtain nonmvdef_names
    nomvdefs = [f.name for f in defs if f.name not in mvdefs]
    nonmvdef_names = get_def_names(nomvdefs, defs, import_annos, extradefs, report)
    if report_VERBOSE:
        print("non-mvdef names:")
        pprint_def_names(nonmvdef_names)
    # ------------------------------------------------------------------------ #
    # Next obtain unused_names
    mv_set = set().union(*[mvdef_names.get(x).keys() for x in mvdef_names])
    nomv_set = set().union(*[nonmvdef_names.get(x).keys() for x in nonmvdef_names])
    unused_names = list(set(list(import_annos[0].keys())) - mv_set - nomv_set)
    nondefs = get_nondef_names(unused_names, import_annos, report=report)
    # undef_names contains only those names that are imported but never used
    undef_names = dict([(x, nondefs.get(x)) for x in nondefs if x not in extradefs])
    if report_VERBOSE:
        rint("non-def names (imported but not used in any function def):")
        pprint_def_names(nondefs, no_funcdef_list=True)
    return mvdef_names, nonmvdef_names, undef_names



from mvdef.__env__ import example_dir
from mvdef.transfer import parse_transfer
from example.test.test_demo import test_report as demotest


def main(mvdefs, dry_run=True, report=True):
    if report:
        print("--------------RUNNING mvdef.demo⠶main()--------------")
    # Step 1: declare src and dst .py file paths and back up the files
    src_p, dst_p = (example_dir / f"{n}.py" for n in ["demo_program", "new_file"])
    src_parsed, dst_parsed = parse_transfer(
        src_p, dst_p, mvdefs=mvdefs, test_func=demotest, report=report, nochange=dry_run
    )
    if report:
        print("------------------COMPLETE--------------------------")
    return



from asttokens import ASTTokens
from ast import Import as IType, ImportFrom as IFType, FunctionDef, walk


def get_tokenised(filepath):
    with open(filepath, "r") as f:
        source = f.read()
    tokenised = ASTTokens(source, parse=True)
    return tokenised


def get_tree(filepath):
    tree = get_tokenised(filepath).tree
    return tree


def get_imports(tr, index_list=None, trunk_only=True):
    """
    Using the `asttoken`-tokenised AST tree body ("trunk"), get the
    top-level import statements. Alternatively, get imports at any
    level by walking the full tree rather than just the trunk.
    """
    if trunk_only:
        imports = [t for t in tr if type(t) in (IType, IFType)]
    else:
        imports = [t for t in walk(tr) if type(t) in (IType, IFType)]
    if index_list is not None:
        for (n, n_i) in index_list:
            return [imports[i] for i in index_list]
    return imports


def get_defs(tr, def_list=[], trunk_only=True):
    """
    Using the `asttoken`-tokenised AST tree body ("trunk"), get the top-level
    function definition statements. Alternatively, get function definitions
    at any level by walking the full tree rather than just the trunk.
    """
    if trunk_only:
        defs = [t for t in tr if type(t) is FunctionDef]
    else:
        defs = [t for t in walk(tr) if type(t) is FunctionDef]
    if def_list == []:
        return defs
    else:
        return [d for d in defs if d.name in def_list]


def locate_import_ends(source_filepath, index_list=None):
    ends = []
    nodes = get_imports(source_filepath, index_list)
    for n in nodes:
        end = {}
        end["line"], end["index"] = n.last_token.end[0]
        ends.append(end)
    return ends



from mvdef.transfer import parse_transfer


def main(src_p, dst_p, mvdefs, dry_run, report, backup):
    if report:
        print("--------------RUNNING mvdef.cli⠶main()--------------")
    src_parsed, dst_parsed = parse_transfer(
        src_p,
        dst_p,
        mvdefs,
        test_func=None,
        report=report,
        nochange=dry_run,
        use_backup=backup,
    )
    if report:
        print("------------------COMPLETE--------------------------")
    return



def pprint_def_names(def_names, no_funcdef_list=False):
    if no_funcdef_list:
        print("  {")
        for m in def_names:
            print(f"    {m}: {def_names.get(m)}")
        print("  }")
    else:
        for n in def_names:
            print(f"  {n}:::" + "{")
            for m in def_names.get(n):
                print(f"    {m}: {def_names.get(n)[m]}")
            print("  }")
    return



import ast
from ast import Import as IType, ImportFrom as IFType
from astor import to_source
from asttokens import ASTTokens
from collections import OrderedDict
from mvdef.colours import colour_str as colour
from os import linesep as nl


def get_import_stmt_str(alias_list, import_src=None, max_linechars=88):
    """
    Construct an import statement by building an AST, convert it to source using
    astor.to_source, and then return the string.

      alias_list:     List of strings to use as ast.alias `name`, and optionally also
                      `asname entries. If only one name is listed per item in the
                      alias_list, the `asname` will be instantiated as None.
      import_src:     If provided, the import statement will be use the
                      `ast.ImportFrom` class, otherwise it will use `ast.Import`.
                      Relative imports are permitted for "import from" statements
                      (such as `from ..foo import bar`) however absolute imports
                      (such as `from foo import bar`) are recommended in PEP8.
      max_linechars:  Maximum linewidth, beyond which the import statement string will
                      be multilined with `multilinify_import_stmt_str`.
    """
    alias_obj_list = []
    assert type(alias_list) is list, "alias_list must be a list"
    for alias_pair in alias_list:
        if type(alias_pair) is str:
            alias_pair = [alias_pair]
        assert len(alias_pair) > 0, "Cannot import using an empty string"
        assert type(alias_pair[0]) is str, "Import alias name must be a string"
        if len(alias_pair) < 2:
            alias_pair.append(None)
        al = ast.alias(*alias_pair[0:2])
        alias_obj_list.append(al)
    if import_src is None:
        ast_imp_stmt = ast.Import(alias_obj_list)
    else:
        import_level = len(import_src) - len(import_src.lstrip("."))
        import_src = import_src.lstrip(".")
        ast_imp_stmt = ast.ImportFrom(import_src, alias_obj_list, level=import_level)
    import_stmt_str = to_source(ast.Module([ast_imp_stmt]))
    if len(import_stmt_str.rstrip(nl)) > max_linechars:
        return multilinify_import_stmt_str(import_stmt_str)
    else:
        return import_stmt_str


def multilinify_import_stmt_str(import_stmt_str, indent_spaces=4, trailing_comma=True):
    """
    Takes a single line import statement and turns it into a multiline string.
    Will raise a `ValueError` if given a multiline string (a newline at the end
    of the string is permitted).

    This function is written in expectation of the output of `get_import_stmt_str`,
    and is not intended to process all potential ways of writing an import statement.

        import_stmt_str:  String of Python code carrying out an import statement.
        indent_spaces:    Number of spaces to indent by in multiline format.
        trailing_comma:   Whether to add a trailing comma to the final alias in a
                          multiline list of import aliases (default: True)
    """
    import_stmt_str = import_stmt_str.rstrip(nl)
    n_nl = import_stmt_str.count(nl)
    if n_nl > 0:
        raise ValueError(f"{import_stmt_str} is not a single line string")
    imp_ast = ast.parse(import_stmt_str)
    assert type(imp_ast.body[0]) in [IType, IFType], "Not a valid import statement"
    tko = ASTTokens(import_stmt_str)
    first_tok = tko.tokens[0]
    import_tok = tko.find_token(first_tok, tok_type=1, tok_str="import")
    assert import_tok.type > 0, f"Unable to find import token in the given string"
    imp_preamble_str = import_stmt_str[: import_tok.endpos]
    post_import_tok = tko.tokens[import_tok.index + 1]
    imp_names_str = import_stmt_str[post_import_tok.startpos :]
    aliases = [(x.name, x.asname) for x in imp_ast.body[0].names]
    seen_comma_tok = None
    multiline_import_stmt_str = imp_preamble_str
    multiline_import_stmt_str += " (" + nl
    for al_i, (a_n, a_as) in enumerate(aliases):
        is_final_alias = al_i + 1 == len(aliases)
        if seen_comma_tok is None:
            # Get start of alias by either full name or first part of .-separated name
            al_n_tok = tko.find_token(import_tok, 1, tok_str=a_n.split(".")[0])
            assert al_n_tok.type > 0, f"Unable to find the token for {a_n}"
        else:
            al_n_tok = tko.find_token(seen_comma_tok, 1, tok_str=a_n.split(".")[0])
            assert al_n_tok.type > 0, f"Unable to find the token for {a_n}"
        al_startpos = al_n_tok.startpos
        if a_as is None:
            if is_final_alias:
                # There won't be a comma after this (it is the last import name token)
                al_endpos = al_n_tok.endpos
            else:
                comma_tok = tko.find_token(al_n_tok, tok_type=53, tok_str=",")
                if comma_tok.type == 0:
                    # Due to an error in asttokens, sometimes tok_type is given as 54
                    # although this should be an error (the failure tok_type is 0)
                    comma_tok = tko.find_token(al_n_tok, tok_type=54, tok_str=",")
                assert comma_tok.type > 0, f"Unable to find comma token"
                al_endpos = comma_tok.endpos
        else:
            al_as_tok = tko.find_token(al_n_tok, tok_type=1, tok_str=a_as)
            assert al_as_tok.type > 0, f"Unable to find the token for {a_as}"
            if is_final_alias:
                # There won't be a comma after this (it's the last import asname token)
                al_endpos = al_as_tok.endpos
            else:
                comma_tok = tko.find_token(al_as_tok, tok_type=53, tok_str=",")
                if comma_tok.type == 0:
                    # Due to an error in asttokens, sometimes tok_type is given as 54
                    # although this should be an error (the failure tok_type is 0)
                    comma_tok = tko.find_token(al_n_tok, tok_type=54, tok_str=",")
                assert comma_tok.type > 0, f"Unable to find comma token"
                al_endpos = comma_tok.endpos
        alias_chunk = import_stmt_str[al_startpos:al_endpos]
        if is_final_alias:
            if trailing_comma:
                alias_chunk += ","
        else:
            seen_comma_tok = comma_tok
        multiline_import_stmt_str += (" " * indent_spaces) + alias_chunk + nl
    # Finally, verify that the end of the tokenised string was reached
    assert al_endpos == tko.tokens[-1].endpos, "Did not tokenise to the end of string"
    # No need to further slice the input string, return the final result
    multiline_import_stmt_str += ")" + nl
    return multiline_import_stmt_str


def colour_imp_stmt(imp_stmt, lines):
    """
    Summary: get a string which when printed will show the separate parts of an
    import statement in different colours (preamble in blue, alias names in red,
    alias asnames in purple, the word "as" itself in yellow, commas between import
    aliases in light green, and post-matter (a bracket) in light red.

    For an import statement within an asttokens-annotated AST, which comes with
    all subnodes annotated with first and last token start/end positional information,
    access all the tokens corresponding to the import statement name(s) and asname(s).
    
    Do this using a list of lines (i.e. a list of strings, each of which is a line),
    the subset of which corresponding to the import statement `imp_stmt` are given
    by its `first_token.start` and `last_token.end` attributes (in each case, the
    attribute is a tuple of `(line, column)` numbers, and it is conventional to store
    these as a 1-based index, so to cross-reference to a 0-based index of the list
    of lines we decrement this value and store as `imp_startln` and `imp_endln`).
    The subset of lines corresponding to `imp_stmt` is then assigned as `nodelines`,
    and we join this into a single string as `nodestring`.

    Then a new ASTTokens object, `tko`, can be made by parsing `nodestring`, on which
    the `find_tokens` method provides access to each name/asname one at a time, when
    matched to the name/asname string. These name/asname strings are available
    within the `imp_stmt` object via its `names` attribute, which is a list of
    `ast.alias` class instances, each of which has both a `name` and `asname` attribute
    (the latter of which is `None` if no asname is given for the import name).

    `find_tokens` returns a token with attribute `type` of value `1` for a name (1 is
    the index of "NAME" in the `token.tok_name` dictionary), and `startpos`/`endpos`
    attributes (integers which indicate the string offsets within `nodestring`).

    These `startpos` integers are an efficient way to store this list of tokens
    (the "NAME" tokens corresponding to import statement alias names and asnames),
    and so even though it would be possible to store all tokens, I choose to simply
    re-access them with the `tko.get_token_from_offset(startpos)` method.

    At the moment, I only re-access these tokens to retrieve their `endpos` (end
    position offset), which is also an integer and could also be stored easily
    without much problem, however for the sake of clarity I prefer to re-access
    the entire token and not have to construct an arbitrary data structure for
    storing the start and end positions (which could get confusing).

    Lastly, I construct a colourful string representation of the import statement
    by using these start positions and re-retrieved end positions to pull out
    and modify (using the `mvdef.colours`⠶`colour_str` function) the names and asnames
    (names are coloured red, asnames are coloured purple), and use string slicing
    to swap the ranges that the names and asnames were in in the original
    `nodestring` for these colourful replacements.

    The end result, `modified_nodestring` is then returned, which will then
    display in colour on Linux and OSX (I don't think Windows supports ANSI codes,
    so I made `colour_str` only apply on these platforms).
    """
    assert "first_token" in imp_stmt.__dir__(), "Not an asttokens-annotated AST node"
    assert type(imp_stmt) in [IType, IFType], "Not an import statement"
    is_from = type(imp_stmt) is IFType
    imp_startln = imp_stmt.first_token.start[0] - 1  # Use 0-based line index
    imp_endln = imp_stmt.last_token.end[0] - 1  # to match list of lines
    nodelines = lines[imp_startln : (imp_endln + 1)]
    n_implines = len(nodelines)
    nodestring = "".join(nodelines)
    tko = ASTTokens(nodestring)
    new_nodelines = [list() for _ in range(n_implines)]
    # Subtract the import statement start position from the name or asname
    # token start position to get the offset, then use the offset to extract
    # a range of text from the re-parsed ASTTokens object for the nodestring
    # corresponding to the import name or asname in question.
    imp_startpos = imp_stmt.first_token.startpos
    alias_starts = []
    for alias in imp_stmt.names:
        al_n, al_as = alias.name, alias.asname
        # 1 is the key for "NAME" in Python's tokens.tok_name
        s = [tko.find_token(tko.tokens[0], 1, tok_str=al_n).startpos]
        if al_as is not None:
            s.append(tko.find_token(tko.tokens[0], 1, tok_str=al_as).startpos)
        alias_starts.append(s)
    assert len(alias_starts) > 0, "An import statement cannot import no names!"
    assert alias_starts[0][0] > 0, "An import statement cannot begin with a name!"
    modified_nodestring = ""
    # -------------------------------------------------------------------------
    # Now set up colour definitions for the modified import statement string
    name_colour, asname_colour = ["red", "purple"]
    pre_colour, post_colour = ["light_blue", "light_red"]
    as_string_colour = "yellow"
    comma_colour = "light_green"
    # -------------------------------------------------------------------------
    first_import_name_startpos = alias_starts[0][0]
    pre_str = nodestring[:first_import_name_startpos]
    modified_nodestring += colour(pre_colour, pre_str)
    seen_endpos = first_import_name_startpos
    # (Could add a try/except here to verify colours are in colour dict if modifiable)
    for al_i, alias_start_list in enumerate(alias_starts):
        for al_j, al_start in enumerate(alias_start_list):
            if seen_endpos < al_start:
                # There is an intervening string, append it to modified_nodestring
                intervening_str = nodestring[seen_endpos:al_start]
                if al_j > 0:
                    # This is the word "as", which comes between a name and an asname
                    modified_nodestring += colour(as_string_colour, intervening_str)
                else:
                    if al_i > 0:
                        assert "," in intervening_str, "Import aliases not comma-sep.?"
                        modified_nodestring += colour(comma_colour, intervening_str)
                    else:
                        modified_nodestring += intervening_str
            # Possible here to distinguish between names and asnames by al_j if needed
            is_asname = bool(al_j)  # al_j is 0 if name, 1 if asname
            name_tok = tko.get_token_from_offset(al_start)
            assert name_tok.type > 0, f"No import name at {al_start} in {nodestring}"
            al_endpos = name_tok.endpos
            imp_name = nodestring[al_start:al_endpos]
            cstr_colour = [name_colour, asname_colour][al_j]
            cstr = colour(cstr_colour, imp_name)
            modified_nodestring += cstr
            seen_endpos = al_endpos
    end_str = nodestring[seen_endpos:]
    modified_nodestring += colour(post_colour, end_str)
    return modified_nodestring


def get_imported_name_sources(trunk, report=True):
    import_types = [IType, IFType]
    imports = [n for n in trunk if type(n) in import_types]
    imp_name_lines, imp_name_dict_list = annotate_imports(imports, report=report)
    imported_names = {}
    for ld in imp_name_lines:
        ld_n = imp_name_lines.get(ld).get("n")
        line_n = imp_name_dict_list[ld_n]
        imp_src = [x for x in list(line_n.items()) if x[1] == ld][0]
        imported_names[ld] = imp_src
    return imported_names


def get_module_srcs(imports):
    ifr_srcs = []
    for imp in imports:
        if type(imp) == IFType:
            ifr_srcs.append(imp.module)
        else:
            ifr_srcs.append(None)
    return ifr_srcs


def count_imported_names(nodes):
    """
    Return an integer for a single node (0 if not an import statement),
    else return a list of integers for a list of AST nodes.
    """
    if type(nodes) is not list:
        if type(nodes) in [IType, IFType]:
            return len(nodes.names)
        else:
            assert ast.stmt in type(nodes).mro(), f"{nodes} is not an AST statement"
            return 0
    counts = []
    for node in nodes:
        if type(node) in [IType, IFType]:
            counts.append(len(node.names))
        else:
            assert ast.stmt in type(nodes).mro(), f"{nodes} is not an AST statement"
            counts.append(0)
    return counts


def annotate_imports(imports, report=True):
    """
    Produce two data structures from the list of import statements (the statements
    of type ast.Import and ast.ImportFrom in the source program's AST),
      imp_name_linedict:  A dictionary whose keys are all the names imported by the
                          program (i.e. the names which they are imported as: the
                          asname if one is used), and whose value for each name
                          is a dictionary of keys (`n`, `line`):
                            n:    [0-based] index of the import statement importing
                                  the name, over the set of all import statements.
                            line: [1-based] line number of the file of the import
                                  statement importing the name. Note that it may
                                  not correspond to the line number on which the
                                  name is given, only to the import function call.
      imp_name_dict_list: List of one OrderedDict per import statement, whose keys
                          are the full import path (with multi-part paths conjoined
                          by a period `.`) and the values of which are the names
                          that these import paths are imported as (either the asname
                          or else just the terminal part of the import path). The
                          OrderedDict preserves the per-line order of the imported
                          names.
    """
    report_VERBOSE = False  # Silencing debug print statements
    # This dictionary gives the import line it's on for cross-ref with either
    # the imports list above or the per-line imported_name_dict
    imp_name_linedict = dict()  # Stores all names and their asnames
    imp_name_dict_list = []  # Stores one OrderedDict per AST import statement
    for imp_no, imp_line in enumerate(imports):
        imp_name_dict = OrderedDict()
        for imported_names in imp_line.names:
            name, asname = imported_names.name, imported_names.asname
            if type(imp_line) == IFType:
                assert imp_line.level == 0, "I've only encountered level 0 imports"
                fullname = ".".join([imp_line.module, name])
            else:
                fullname = name
            if asname is None:
                imp_name_dict[fullname] = name
                # Store both which import in the list of imports it's in
                # and the line number it's found on in the parsed file
                imp_name_linedict[name] = {"n": imp_no, "line": imp_line.lineno}
            else:
                imp_name_dict[fullname] = asname
                imp_name_linedict[asname] = {"n": imp_no, "line": imp_line.lineno}
        imp_name_dict_list.append(imp_name_dict)
    # Ensure that they each got all the names
    assert len(imp_name_dict_list) == len(imports)
    assert sum([len(d) for d in imp_name_dict_list]) == len(imp_name_linedict)
    if report_VERBOSE:
        print("The import name line dict is:")
        for ld in imp_name_linedict:
            # print(f"  {ld}: {imp_name_linedict.get(ld)}")
            pass
        print("The import name dict list is:")
        for ln in imp_name_dict_list:
            print(ln)
    return imp_name_linedict, imp_name_dict_list


def imp_def_subsets(mvdefs, nonmvdefs, report=True):
    """
    Given the list of mvdef_names and nonmvdef_names, construct the subsets:
      mv_imports:      imported names used by the functions to move,
      nonmv_imports:   imported names used by the functions not to move,
      mutual_imports:  imported names used by both the functions to move and
                        the functions not to move
    """
    report_VERBOSE = False  # Silencing debug print statements
    mvdefs_names = set().union(*[list(mvdefs[x]) for x in mvdefs])
    nonmvdefs_names = set().union(*[list(nonmvdefs[x]) for x in nonmvdefs])
    mv_imports = mvdefs_names - nonmvdefs_names
    nonmv_imports = nonmvdefs_names - mvdefs_names
    mutual_imports = mvdefs_names.intersection(nonmvdefs_names)
    assert mv_imports.isdisjoint(nonmv_imports), "mv/nonmv_imports intersect!"
    assert mv_imports.isdisjoint(mutual_imports), "mv/mutual imports intersect!"
    assert nonmv_imports.isdisjoint(mutual_imports), "nonmv/mutual imports intersect!"
    if report_VERBOSE:
        print(
            f"mv_imports: {mv_imports}",
            f", nonmv_imports: {nonmv_imports}",
            f", mutual_imports: {mutual_imports}",
            sep="",
        )
    all_defnames = set().union(*[mvdefs_names, nonmvdefs_names])
    all_def_imports = set().union(*[mv_imports, nonmv_imports, mutual_imports])
    assert sorted(all_defnames) == sorted(all_def_imports), "Defnames =/= import names"
    return mv_imports, nonmv_imports, mutual_imports



def backup(filepath, dry_run=False, suffix=".backup", hidden=True):
    """
    Given a filename, copy it to a backup before making changes and confirm success to
    allow an in place edit to proceed safely. If a file with the same name exists, add
    an incrementing integer. If dry_run is True, all the checks for the possibility of
    creating the backup will be run, but no files will be opened or touched. (To report
    on the funcdef/import statements to be moved without having to adhere to these
    requirements, don't call `mvdef.backup.backup` before calling `mvdef.ast_util.ast_parse`).
    """
    fd = filepath.parent
    assert fd.exists() and fd.is_dir(), f"Can't backup {filepath}: {fd} doesn't exist"
    hid_prefix = "." * int(hidden)  # Empty string if hidden is False
    if not filepath.exists():
        # This file is a dst to be created, make an empty backup (indicates no restore)
        assert not filepath.exists()
        empty_msg = f"# EMPTY BACKUP FOR {filepath.name} CREATED BY `mvdef⠶backup()`"
        b_file = fd / f"{hid_prefix}{filepath.name}.backup"
        # Do not tolerate even a single backup file for a file to be newly created
        assert not b_file.exists(), f"Backup {b_file} exists for non-existing file!"
        if not dry_run:
            with open(b_file, "w") as f:
                f.write(f"{empty_msg}\n")
        return True
    else:
        assert filepath.exists() and filepath.is_file() and filepath.suffix == ".py"
    bname = f"{hid_prefix}{filepath.name}{suffix}"
    if fd / bname in fd.iterdir():
        for i in range(0, 12):
            bname_i = fd / f"{bname}{i}"
            if bname_i not in fd.iterdir():
                break
            i += 1
            if i > 10:
                raise ValueError("There are over 10 backups, something's wrong")
        assert not (fd / bname_i).exists()
        # Use the filename {bname_i} for the backup
        bname = bname_i
    assert not (fd / bname).exists()
    if not dry_run:
        with open(filepath, "r") as o:
            original = o.read()
        with open(fd / bname, "w") as b:
            b.write(original)
    return True



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()
 
setup(
    name="mvdef",
    version="0.2.8",
    author="Louis Maddox",
    author_email="louismmx@gmail.com",
    description="Package to move functions and their import statements between files",
    license="MIT License",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lmmx/mvdef",
    packages=find_packages('src'),
    package_dir={'': 'src'},
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
    ],
    python_requires='>=3',
)



from collections import OrderedDict
from mvdef.ast_tokens import get_defs, get_imports, get_tree
from mvdef.ast_util import annotate_imports
from mvdef.editor_util import get_def_lines, get_defrange, excise_def_lines, overwrite_import
from mvdef.import_util import get_import_stmt_str, get_module_srcs, count_imported_names


def transfer_mvdefs(src_path, dst_path, mvdefs, src_agenda, dst_agenda):
    # Firstly annotate ASTs with the asttokens library
    with open(src_path, "r") as f:
        src_lines = f.readlines()
    with open(dst_path, "r") as f:
        dst_lines = f.readlines()
    src_trunk = get_tree(src_path).body
    dst_trunk = get_tree(dst_path).body
    src_mvdefs = get_defs(tr=src_trunk, def_list=mvdefs, trunk_only=True)
    src_imports = get_imports(src_trunk, trunk_only=True)
    dst_imports = []
    dst_imports = get_imports(dst_trunk, trunk_only=True)
    src_import_counts = count_imported_names(src_imports)
    dst_import_counts = count_imported_names(dst_imports)
    src_modules = get_module_srcs(src_imports)
    dst_modules = get_module_srcs(dst_imports)
    # ------------------------- First move the imports ------------------------------
    # Do not need to handle "copy", "keep", or "stay" edit_agenda entries,
    # "copy" entries in src_agenda are mirrored by "echo" entries in dst_agenda,
    # while all "move" entries in src_agenda are mirrored as "take" in dst_agenda
    # -------------------------------------------------------------------------------
    # Convert lose list of info dicts into OrderedDict of to-be-removed names/info
    dst_rm_agenda = OrderedDict([list(a.items())[0] for a in dst_agenda.get("lose")])
    # Merge take/echo lists of info dicts into OrderedDict of received names/info
    dst_rcv_agenda = dst_agenda.get("take") + dst_agenda.get("echo")
    dst_rcv_agenda = OrderedDict([list(a.items())[0] for a in (dst_rcv_agenda)])
    # Merge lose/move lists of info dicts into OrderedDict of to-be-removed names/info
    src_rm_agenda = src_agenda.get("move") + src_agenda.get("lose")
    src_rm_agenda = OrderedDict([list(a.items())[0] for a in (src_rm_agenda)])
    #
    # ----------------- STEP 1: REMOVE IMPORTS MARKED DST⠶LOSE ----------------------
    #
    removed_import_n = []
    for rm_i in dst_rm_agenda:
        # Remove rm_i (imported name marked "lose") from the destination file using
        # the line numbers of `dst_trunk`, computed as `dst_imports` by `get_imports`
        # (a destructive operation, so line numbers of `dst_trunk` no longer valid),
        # if the removal of the imported name leaves no other imports on a line,
        # otherwise shorten that line by removing the import alias(es) marked "lose"
        dst_info = dst_rm_agenda.get(rm_i)
        imp_src_ending = dst_info.get("import").split(".")[-1]
        # Retrieve the index of the line in import list
        rm_i_n = dst_info.get("n")
        rm_i_linecount = dst_import_counts[rm_i_n]
        if rm_i_linecount > 1:
            # This means there is ≥1 other import alias in the import statement
            # for this name, so remove it from it (i.e. "shorten" the statement.
            dst_info["shorten"] = dst_info.get("n_i")
        else:
            # This means there is nothing from this module being imported yet, so
            # must remove entire import statement (i.e. delete entire line range)
            removed_import_n.append(rm_i_n)
            imp_startline = dst_imports[rm_i_n].first_token.start[0]
            imp_endline = dst_imports[rm_i_n].last_token.end[0]
            imp_linerange = [imp_startline - 1, imp_endline]
            for i in range(*imp_linerange):
                dst_lines[i] = None
            dst_info["shorten"] = None
    to_shorten = []
    for rm_i in dst_rm_agenda:
        if dst_rm_agenda.get(rm_i).get("shorten") is not None:
            to_shorten.append((rm_i, dst_rm_agenda.get(rm_i)))
    to_shorten = OrderedDict(to_shorten)
    n_to_short = set([to_shorten.get(x).get("n") for x in to_shorten])
    # Group all names being shortened that are of a common import statement
    for n in n_to_short:
        names_to_short = [x for x in to_shorten if to_shorten.get(x).get("n") == n]
        n_i_to_short = [to_shorten.get(a).get("n_i") for a in names_to_short]
        # Rewrite `dst_imports[n]` with all aliases except those in `names_to_short`
        imp_module = dst_modules[n]
        pre_imp = dst_imports[n]
        shortened_alias_list = [(a.name, a.asname) for a in pre_imp.names]
        # Proceed backwards from the end to the start, permitting deletions by index
        for (name, asname) in shortened_alias_list[::-1]:
            if asname is None and name not in names_to_short:
                continue
            elif asname not in names_to_short:
                continue
            del_i = shortened_alias_list.index((name, asname))
            del shortened_alias_list[del_i]
        if len(shortened_alias_list) == 0:
            # All import aliases were removed, so remove the entire import statement
            removed_import_n.append(n)
            imp_startline = pre_imp.first_token.start[0]
            imp_endline = pre_imp.last_token.end[0]
            imp_linerange = [imp_startline - 1, imp_endline]
            for i in range(*imp_linerange):
                dst_lines[i] = None
        else:
            imp_stmt_str = get_import_stmt_str(shortened_alias_list, imp_module)
            overwrite_import(pre_imp, imp_stmt_str, dst_lines)
    #
    # --------------- STEP 2: ADD IMPORTS MARKED DST⠶{MOVE,COPY} --------------------
    #
    for rc_i in dst_rcv_agenda:
        # Transfer mv_i into the destination file: receive "move" as "take"
        # Transfer cp_i into the destination file: receive "copy" as "echo"
        dst_info = dst_rcv_agenda.get(rc_i)
        imp_src_ending = dst_info.get("import").split(".")[-1]
        # Use name/asname to retrieve the index of the line in import list to get
        # the module which is at the same index in the list of src modules:
        rc_i_n = dst_info.get("n")
        rc_i_module = src_modules[rc_i_n]
        # Compare the imported name to the module if one exists
        if rc_i_module is not None:
            dst_module_set = set(dst_modules).difference({None})
            if rc_i_module in dst_module_set:
                # This means there is already ≥1 ast.ImportFrom statement (i.e. a
                # line) which imports from the same module as the to-be-added import
                # name does, so combine it with this existing line. Assume the 1st
                # such ImportFrom is to be extended (ignoring other possible ones).
                dst_info["extend"] = dst_modules.index(rc_i_module)
            else:
                # This means there is nothing from this module being imported yet,
                # so must create a new ImportFrom statement (i.e. a new line)
                dst_info["extend"] = None
        else:
            # This means `rc_i` is an ast.Import statement, not ImportFrom
            # (PEP8 recommends separate imports, so do not extend another)
            dst_info["extend"] = None
    to_extend = []
    for rc_i in dst_rcv_agenda:
        if dst_rcv_agenda.get(rc_i).get("extend") is not None:
            to_extend.append((rc_i, dst_rcv_agenda.get(rc_i)))
    to_extend = OrderedDict(to_extend)
    n_to_extend = set([to_extend.get(x).get("n") for x in to_extend])
    # Group all names being added as extensions that are of a common import statement
    for n in n_to_extend:
        names_to_extend = [x for x in to_extend if to_extend.get(x).get("n") == n]
        # Rewrite `dst_imports[n]` to include the aliases in `names_to_extend`
        imp_module = dst_modules[n]
        pre_imp = dst_imports[n]
        extended_alias_list = [(a.name, a.asname) for a in pre_imp.names]
        for rc_i in names_to_extend:
            dst_info = to_extend.get(rc_i)
            imp_src = dst_info.get("import")
            imp_src_ending = imp_src.split(".")[-1]
            if rc_i == imp_src_ending:
                rc_i_name, rc_i_as = rc_i, None
            elif imp_module is not None:
                rc_i_name, rc_i_as = imp_src_ending, rc_i
            else:
                rc_i_name, rc_i_as = imp_src, rc_i
            extended_alias_list.append((rc_i_name, rc_i_as))
        imp_stmt_str = get_import_stmt_str(extended_alias_list, imp_module)
        overwrite_import(pre_imp, imp_stmt_str, dst_lines)
    # Next, put any import names marked "take" or "echo" that are not extensions
    # of existing import statements into new lines (this breaks the line index).
    #
    # Firstly, find the insertion point for new import statements by re-processing
    # the list of lines (default to start of file if it has no import statements)
    import_n = [n for n, _ in enumerate(dst_imports) if n not in removed_import_n]
    if len(import_n) == 0:
        # Place any new import statements at the start of the file, as none exist yet
        last_imp_end = 0  # 1-based index logic: this means "before the first line"
    else:
        last_import = dst_imports[import_n[-1]]
        last_imp_end = last_import.last_token.end[0]  # Leave in 1-based index
    ins_imp_stmts = []  # Collect import statements to insert after the last one
    seen_multimodule_imports = set()
    for rc_i in dst_rcv_agenda:
        dst_info = dst_rcv_agenda.get(rc_i)
        if rc_i in seen_multimodule_imports or dst_info.get("extend") is not None:
            continue
        imp_src = dst_info.get("import")
        imp_src_ending = imp_src.split(".")[-1]
        rc_i_n = dst_info.get("n")
        rc_i_module = src_modules[rc_i_n]
        if rc_i == imp_src_ending:
            rc_i_name, rc_i_as = rc_i, None
        elif rc_i_module is not None:
            rc_i_name, rc_i_as = imp_src_ending, rc_i
        else:
            rc_i_name, rc_i_as = imp_src, rc_i
        alias_list = [(rc_i_name, rc_i_as)]
        for r in dst_rcv_agenda:
            r_src_module = src_modules[dst_rcv_agenda.get(r).get("n")]
            if r == rc_i or None in [rc_i_module, r_src_module]: continue
            if r_src_module == rc_i_module:
                seen_multimodule_imports.add(r)
                r_dst_info = dst_rcv_agenda.get(r)
                r_imp_src = r_dst_info.get("import")
                r_imp_src_ending = r_imp_src.split(".")[-1]
                r_n = r_dst_info.get("n")
                r_module = src_modules[r_n]
                if r == r_imp_src_ending:
                    r_name, r_as = r, None
                elif r_module is not None:
                    r_name, r_as = r_imp_src_ending, r
                else:
                    r_name, r_as = r_dst_info.get("import"), r
                alias_list.append((rc_i_name, rc_i_as))
        # Create the Import or ImportFrom statement
        imp_stmt_str = get_import_stmt_str(alias_list, rc_i_module)
        ins_imp_stmts.append(imp_stmt_str)
    dst_lines = dst_lines[:last_imp_end] + ins_imp_stmts + dst_lines[last_imp_end:]
    # ------------------------------------------------------------------------------
    # Postpone the extension/addition of import statements (do all at once so as to
    # retain meaningful line numbers, as changing one at a time would ruin index)
    # as well as to minimise file handle opening/reopening since that IO takes time
    # Also: sort the import statements by n in `src_imports` index to retain order
    # ------------------------------------------------------------------------------
    #
    # --------- STEP 3: COPY FUNCTION DEFINITIONS {MVDEFS} FROM SRC TO DST ---------
    #
    for mvdef in src_mvdefs:
        # Transfer mvdef into the destination file: receive mvdef
        # mvdef is an ast.FunctionDefinition node with start/end position annotations
        # using the line numbers of `src_trunk`, computed as `src_mvdefs` by `get_defs`
        # (this is an append operation, so line numbers from `src_trunk` remain valid)
        def_startline, def_endline = get_defrange(mvdef)
        deflines = src_lines[def_startline:def_endline]
        dst_lines += get_def_lines(deflines, dst_lines)
    # -------- Line number preservation no longer needed, only now modify src -------
    # Iterate through funcdefs in reverse line number order (i.e. upward from bottom)
    # using the line numbers of `src_trunk`, computed as `src_mvdefs` by `get_defs`
    #
    # --------- STEP 4: REMOVE FUNCTION DEFINITIONS {MVDEFS} FROM SRC ---------
    #
    for mvdef in sorted(src_mvdefs, key=lambda d: d.last_token.end[0], reverse=True):
        # Remove mvdef (function def. marked "mvdef") from the source file
        excise_def_lines(mvdef, src_lines)
    #
    # --------------- STEP 5: REMOVE IMPORTS MARKED SRC⠶{MOVE,LOSE} ----------------
    #
    for rm_i in src_rm_agenda:
        # Remove rm_i (imported name marked "move"/"lose") from the source file using
        # the line numbers of `src_trunk`, computed as `src_imports` by `get_imports`
        # (a destructive operation, so line numbers of `src_trunk` no longer valid),
        # if the removal of the imported name leaves no other imports on a line,
        # otherwise shorten that line by removing the import alias(es) marked "lose"
        src_info = src_rm_agenda.get(rm_i)
        imp_src_ending = src_info.get("import").split(".")[-1]
        # Retrieve the index of the line in import list
        rm_i_n = src_info.get("n")
        rm_i_linecount = src_import_counts[rm_i_n]
        if rm_i_linecount > 1:
            # This means there is ≥1 other import alias in the import statement
            # for this name, so remove it from it (i.e. "shorten" the statement.
            src_info["shorten"] = src_info.get("n_i")
        else:
            # This means there is nothing from this module being imported yet, so
            # must remove entire import statement (i.e. delete entire line range)
            src_info["shorten"] = None
            imp_startline = src_imports[rm_i_n].first_token.start[0]
            imp_endline = src_imports[rm_i_n].last_token.end[0]
            imp_linerange = [imp_startline - 1, imp_endline]
            for i in range(*imp_linerange):
                src_lines[i] = None
            src_info["shorten"] = None
    to_shorten = []
    for rm_i in src_rm_agenda:
        if src_rm_agenda.get(rm_i).get("shorten") is not None:
            to_shorten.append((rm_i, src_rm_agenda.get(rm_i)))
    to_shorten = OrderedDict(to_shorten)
    n_to_short = set([to_shorten.get(x).get("n") for x in to_shorten])
    # Group all names being shortened that are of a common import statement
    for n in n_to_short:
        names_to_short = [x for x in to_shorten if to_shorten.get(x).get("n") == n]
        n_i_to_short = [to_shorten.get(a).get("n_i") for a in names_to_short]
        # Rewrite `src_imports[n]` with all aliases except those in `names_to_short`
        imp_module = src_modules[n]
        pre_imp = src_imports[n]
        shortened_alias_list = [(a.name, a.asname) for a in pre_imp.names]
        # Proceed backwards from the end to the start, permitting deletions by index
        for (name, asname) in shortened_alias_list[::-1]:
            if asname is None and name not in names_to_short:
                continue
            elif asname is not None and asname not in names_to_short:
                continue
            del_i = shortened_alias_list.index((name, asname))
            del shortened_alias_list[del_i]
        if len(shortened_alias_list) == 0:
            # All import aliases were removed, so remove the entire import statement
            imp_startline = pre_imp.first_token.start[0]
            imp_endline = pre_imp.last_token.end[0]
            imp_linerange = [imp_startline - 1, imp_endline]
            for i in range(*imp_linerange):
                src_lines[i] = None
        else:
            imp_stmt_str = get_import_stmt_str(shortened_alias_list, imp_module)
            overwrite_import(pre_imp, imp_stmt_str, src_lines)
    # Finish by writing line changes back to file (only if agenda shows edits made)
    if len(src_rm_agenda) > 0:
        src_lines = "".join([line for line in src_lines if line is not None])
        with open(src_path, "w") as f:
            f.write(src_lines)
    if len(dst_rcv_agenda) + len(dst_rm_agenda) > 0:
        dst_lines = "".join([line for line in dst_lines if line is not None])
        with open(dst_path, "w") as f:
            f.write(dst_lines)
    return



from mvdef.ast_util import ast_parse
from mvdef.backup import backup
from mvdef.colours import colour_str as colour
from mvdef.editor import transfer_mvdefs

# TODO: Move parse_example to AST once logic is figured out for the demo
def parse_transfer(
    src_p, dst_p, mvdefs, test_func=None, report=True, nochange=True, use_backup=True
):
    """
    Execute the transfer of function definitions and import statements, optionally
    (if test_func is specified) also calls that afterwards to confirm functionality
    remains intact.

    If test_func is specified, it must only use AssertionError (i.e. you are free
    to have the test_func call other functions, but it must catch any errors therein
    and only raise errors from assert statements). This is to simplify this step.
    My example would be to list one or more failing tests, then assert that this
    list is None, else raise an AssertionError of these tests' definition names
    (see example.test.test_demo⠶test_report for an example of such a function).

    If nochange is False, files will be changed in place (i.e. setting it
    to False is equivalent to setting the edit parameter to True).
      - Note: this was unimplemented...

    This parameter is used as a sanity check to prevent wasted computation,
    as if neither report is True nor nochange is False, there is nothing to do.
    """
    # Backs up source and target to a hidden file, restorable in case of error,
    # and creating a hidden placeholder if the target doesn't exist yet
    assert True in [report, not nochange], "Nothing to do"
    if test_func is not None:
        try:
            test_func.__call__()
        except AssertionError as e:
            raise RuntimeError(f"! {test_func} failed, aborting mvdef execution.")
    if use_backup:
        assert backup(src_p, dry_run=nochange)
        assert backup(dst_p, dry_run=nochange)
    # Create edit agendas from the parsed AST of source and destination files
    src_edits = ast_parse(src_p, mvdefs=mvdefs, report=report)
    assert src_edits is not None, "The src file did not return a processed AST"
    if report:
        print(f"⇒ Functions moving from {colour('light_gray', src_p)}: {mvdefs}")
    transfers = dict([["take", src_edits.get("move")], ["echo", src_edits.get("copy")]])
    # Create the destination file if it doesn't exist, and if this isn't a dry run
    dst_extant = dst_p.exists() and dst_p.is_file()
    if not dst_extant and not nochange:
        with open(dst_p, "w") as f:
            f.write("")
    dst_edits = ast_parse(dst_p, transfers=transfers, report=report)
    if dst_edits is None:
        # There is no destination file (it will be created)
        if report:
            print(
                f"⇒ Functions will move to {colour('light_gray', dst_p)}"
                + " (it's being created from them)"
            )
    else:
        if report:
            print(f"⇒ Functions will move to {colour('light_gray', dst_p)}")
    if nochange:
        print("DRY RUN: No files have been modified, skipping tests.")
        return src_edits, dst_edits
    else:
        # Edit the files (no longer pass imports or defs, will recompute AST)
        transfer_mvdefs(src_p, dst_p, mvdefs, src_edits, dst_edits)
    if test_func is None:
        return src_edits, dst_edits
    else:
        try:
            test_func.__call__()
        except AssertionError as e:
            # TODO: implement backup restore
            print(
                f"! {test_func} failed, indicating changes made by mvdef broke the"
                + "program (if backups used, mvdefs will now attempt to restore)"
            )
            raise RuntimeError(e)
    return src_edits, dst_edits



from platform import system as platform


def get_colour_codes(colour=None):
    """
    Return a dictionary of ANSI colour codes if no colour specified,
    if a colour is given then check it is in the list of keys and 
    return a start and end ANSI code.
    """
    cols = [
        ["black", "\033[0;30m"],
        ["blue", "\033[0;34m"],
        ["brown", "\033[0;33m"],
        ["cyan", "\033[0;36m"],
        ["dark_gray", "\033[1;30m"],
        ["green", "\033[0;32m"],
        ["light_blue", "\033[1;34m"],
        ["light_cyan", "\033[1;36m"],
        ["light_gray", "\033[0;37m"],
        ["light_green", "\033[1;32m"],
        ["light_purple", "\033[1;35m"],
        ["light_red", "\033[1;31m"],
        ["light_white", "\033[1;37m"],
        ["purple", "\033[0;35m"],
        ["red", "\033[0;31m"],
        ["yellow", "\033[1;33m"],
    ]
    ending = "\033[0m"
    colour_dict = dict([[x[0], {"on": x[1], "off": ending}] for x in sorted(cols)])
    if colour is None:
        return colour_dict
    else:
        assert colour in colour_dict, f"{colour} not in {list(colour_dict.keys())}"
        return colour_dict.get(colour)


def get_effect_codes(effect=None):
    """
    Return a dictionary of ANSI text effect codes if no effect specified,
    if a text effect is given then check it is in the list of keys and return
    a start and end ANSI code.
    """
    effects = [
        ["bold", "\033[1m"],
        ["faint", "\033[2m"],
        ["italic", "\033[3m"],
        ["underline", "\033[4m"],
        ["blink", "\033[5m"],
        ["negative", "\033[7m"],
        ["crossed", "\033[9m"],
    ]
    ending = "\033[0m"
    effect_dict = dict([[x[0], {"on": x[1], "off": ending}] for x in sorted(effects)])
    if effect is None:
        return effect_dict
    else:
        assert effect in effect_dict, f"{effect} not in {list(effect_dict.keys())}"
        return effect_dict.get(effect)


def colour_str(colour, text, end=True):
    """
    Apply the ANSI colour code to a string, and optionally append the end code.
    Only use `end=False` when combining ANSI codes (i.e. in which case the end
    code would apply to all modifiers so as to prevent their combination).
    """
    if platform().lower() not in ["linux", "darwin"]:
        return text
    colour_on, colour_off = get_colour_codes(colour).values()
    colourful = f'{colour_on}{text}{colour_off if end else ""}'
    return colourful


def effect_str(effect, text, end=True):
    """
    Apply the ANSI effect code to a string, and optionally append the end code.
    Only use `end=False` when combining ANSI codes (i.e. in which case the end
    code would apply to all modifiers so as to prevent their combination).
    """
    if platform().lower() not in ["linux", "darwin"]:
        return text
    effect_on, effect_off = get_effect_codes(effect).values()
    effectful = f'{effect_on}{text}{effect_off if end else ""}'
    return effectful


def underline(text, end=True):
    """
    Apply the underline effect ANSI code to a text string.
    """
    return effect_str("underline", text, end)


def colour_effect_str(colour, effect, text):
    """
    Apply multiple ANSI colour codes to a string, ending only the outer one,
    e.g. to both underline and colour, or underline and bold, etc.

    The effect must be applied after a colour code, otherwise it will be
    'overridden' by the colour code (as far as I have tested).

    Note that multiple effects seem to cancel out (with the latter ANSI code
    taking priority, 'overriding' the other), i.e. multi-effect not possible.
    """
    if platform().lower() not in ["linux", "darwin"]:
        return text
    effectful = effect_str(effect, text, end=False)
    combo = colour_str(colour, effectful)
    return combo



from sys import path as syspath
from sys import argv
from pathlib import Path
from argparse import ArgumentParser

# Put the absolute path to the module directory on the system PATH:
syspath.insert(0, str(Path(__file__).parent))

from mvdef.demo import main as run_demo
from mvdef.cli import main as run_cli

parser = ArgumentParser(
    description="Move function definitions and associated import"
    + " statements from one file to another within a library."
)
parser.add_argument("--demo", action="store_true")
parser.add_argument("-m", "--mvdef", action="append")
parser.add_argument("--src", action="store")
parser.add_argument("--dst", action="store")
parser.add_argument("-r", "--report", action="store_true")
parser.add_argument("-b", "--backup", action="store_true")
parser.add_argument("-d", "--dry-run", action="store_true")


if __name__ == "__main__":
    if argv[0].endswith("__main__.py"):
        argv = [a for a in argv if a != argv[0]]
arg_l = parser.parse_args(argv)

if not arg_l.demo:
    pass
    #src_path = Path(arg_l.src).absolute()
    #dst_path = Path(arg_l.dst).absolute()
    #mvdefs = arg_l.mvdef
    #dry_run = arg_l.dry_run
    #report = arg_l.report
    #backup = arg_l.backup
    #run_cli(src_path, dst_path, mvdefs, dry_run, report, backup)
else:
    run_demo(mvdefs=["show_line"], dry_run=False, report=True)



from mvdef.io_util import terminal_whitespace
from numpy import where
from os import linesep as nl


def get_defrange(def_node):
    """
    def_node:    An ast.FunctionDefinition node with start/end position annotations
                 from the asttokens library.
    """
    def_startline = def_node.first_token.start[0] - 1  # Subtract 1 to get 0-base index
    def_endline = def_node.last_token.end[0]  # Don't subtract 1, to include full range
    defrange = [def_startline, def_endline]
    return defrange


def get_defstring(def_node, file_lines):
    """
    def_node:    An ast.FunctionDefinition node with start/end position annotations
                 from the asttokens library.
    file_lines:  A list from readlines (should contain the appropriate file system
                 newlines, the list will be joined with a blank string).
    """
    def_startline, def_endline = get_defrange(def_node)
    deflines = file_lines[def_startline:def_endline]
    defstring = "".join(deflines)
    return defstring


def append_def_to_file(defstring, dst_path):
    """
    Insert the lines of a function defintion into a file (in future this function may
    permit insertion at a certain order position in the file's function definitions).

    `defstring` must be a string containing appropriate newlines for a Python file.
    """
    # Assess the whitespace, leave at least 2
    end_blanklines = terminal_whitespace(dst_path)
    append_newlines = max((0, 2 - end_blanklines)) * nl
    with open(dst_path, "a") as f:
        f.write(append_newlines + defstring)
    return


def get_def_lines(deflines, dst_lines):
    """
    Get the list of lines of a func. def. suitable to be appended to a set of lines.

    `deflines` must be a list of strings containing appropriate newlines for a Python
    file (the lines will not be joined with newlines, they must be already supplied).
    """
    # Assess the whitespace, leave at least 2
    end_blanklines = terminal_whitespace(dst_lines, from_file=False)
    append_newlines = [nl for _ in range(max((0, 2 - end_blanklines)))]
    return append_newlines + deflines


def excise_def_from_file(def_node, py_path, return_def=True):
    """
    Either cut or delete a function definition using its AST node (via asttokens).

    If `return_def` is True, modify the file at `py_path` to remove the lines from
    `def_node.first_token.start[0]` to `def_node.last_token.end[0]`, and return the
    string read from the lines from `py_path` which contained it (i.e. a "cut" operation).
    
    If `return_def` is False, modify the file at `py_path` to remove the lines that
    contain the function called `def_name`, return `None` (i.e. a delete operation).
    """
    with open(py_path, "w+") as f:
        lines = f.readlines()
        # Inkeeping with convention, range is inclusive at start, exclusive at end i.e. [)
        def_startline = def_node.first_token.start[0] - 1
        def_endline = def_node.last_token.end[0]
        defrange = [def_startline, def_endline]
        pre = (def_startline - 2, def_startline)
        post = (def_endline, def_endline + 2)
        # Count whitespace above and below the function definition
        wspace_a = [x == nl for x in lines[pre[0] : pre[1]]]
        wspace_b = [x == nl for x in lines[post[0] : post[1]]]
        wspace_count = (wspace_a + wspace_b).count(True)
        # wspace_added = "" # Don't think I actually need to implement this
        if wspace_count > 2:
            # Remove whitespace: get list of indexes of lines which are blank
            prepost = [p for p in range(*pre)] + [p for p in range(*post)]
            ws_li = [prepost[i] for (i, x) in enumerate((ws_a + ws_b)) if x]
            # Take as many as reduce the whitespace count to 2
            remove_li = [ws_li[n] for n in range(wspace_count - 2)]
            for li in remove_li:
                if li < min(defrange):
                    defrange[0] = li
                elif li > max(defrange):
                    defrange[1] = li
                # Otherwise li is intermediate (already processed a past li, continue)
        # Whitespace count of less than 2 could only happen when a def is at the end of
        # a file, in which case no need to add whitespace, so no need to check for it.
        excised_lines = lines.copy()
        for i in reversed(range(*defrange)):
            del excised_lines[i]  # Delete lines backwards from end of file line range
    if return_def:
        deflines = lines[def_startline:def_endline]
        return deflines
    else:
        # Edit file in place (N.B. will not use this actually)
        return


def excise_def_lines(def_node, lines):
    """
    Delete a function definition using its AST node (via asttokens) from a list of lines
    which originated from a single, entire, Python file.
    """
    def_startline = def_node.first_token.start[0]
    def_endline = def_node.last_token.end[0]
    # Subtract 1 from start line index for 0-based "inclusive start/exclusive end"
    defrange = [def_startline - 1, def_endline]
    pre, post = get_borders(defrange, lines, window_size=2)
    # Count whitespace above and below the function definition
    # Reverse the order of `pre` otherwise redefining the range start to be the 1st
    # would also necessarily include the 2nd elem of `pre` within the range
    wspace_pre = [lines[p] == nl for p in pre[::-1]]
    wspace_post = [lines[p] == nl for p in post]
    wspace_count = (wspace_pre + wspace_post).count(True)
    if wspace_count > 2:
        # Remove whitespace: get list of indexes of lines which are blank
        # Reverse pre so as to match the index of `wspace_pre` as above
        pp = pre[::-1] + post
        ws_li = [pp[i] for (i, x) in enumerate((wspace_pre + wspace_post)) if x]
        # Take as many as reduce the whitespace count to 2
        remove_li = [ws_li[n] for n in range(wspace_count - 2)]
        for li in remove_li:
            # Reduce whitespace border by extending defrange to include it
            if li < min(defrange):
                defrange[0] = li
            elif li > max(defrange):
                defrange[1] = li
            # Otherwise li is intermediate (already processed a past li, continue)
    # Whitespace count of less than 2 could only happen when a def is at the end of
    # a file, in which case no need to add whitespace, so no need to check for it.
    for i in range(*defrange):
        lines[i] = None  # Mark lines as deleted by setting the string to `None`
    return


def get_borders(defrange, lines, window_size=2):
    """
    Given the range corresponding to a function definition, and the list of lines
    this range is in reference to, and the "window size" of the number of lines
    to compare on each [out]side of this range, return the list of indices of
    lines which are not `None`.

    This is necessary when some lines have been 'nullified' by replacing the string
    at that index with `None`, so as to conserve the line numbering while removing
    lines (i.e. after excising import names and/or function definitions).

    The expected value of the range is to follow Python's convention for ranges,
    i.e. "inclusive start, exclusive end" - mathematically written as `[)`.

    Will return two lists of integers which represent the line indexes of the pre-
    and post-function definition non-`None` lines (i.e. the lines above and below,
    ignoring any lines which have previously been removed). If the `window_size`
    of lines above and/or below is not found, the maximum number of lines will be
    given (i.e. returns empty lists if nothing is found above/below the `defrange`).
    """
    d_start, d_end = defrange
    # Get non-`None` line indexes to a max. of `window_size` away from the start
    pre_i = where([l is not None for l in lines[:d_start][::-1]])[0][:window_size]
    # Reverse list of index offset from d_start back to normal order, get abs. index
    pre = d_start - 1 - pre_i[::-1]
    post_i = where([l is not None for l in lines[d_end:]])[0][:window_size]
    post = d_end + post_i
    return pre.tolist(), post.tolist()


def overwrite_import(imp_node, replacement_str, lines):
    """
    Similar to `excise_def_from_lines`, this function overwrites the line or lines
    which correspond to an import statement (the AST node of which is given as
    `imp_node`), either replacing each line with the equivalent of `replacement_str`
    (the replacement import statement string generated by `get_import_stmt_str`)
    if the replacement is the same number of lines long, or if the number of lines
    has changed, it will not split the lines and simply place the entire replacement
    import statement string in one entry of the `lines` list, and set the "spare"
    entries over the range previously occupied by the import statement to `None`.
    """
    if not replacement_str.endswith(nl):
        replacement_str += nl
    pre_startline = imp_node.first_token.start[0] - 1
    pre_endline = imp_node.last_token.end[0]
    len_pre = pre_endline - pre_startline + 2
    replacement_lines = [f"{r}{nl}" for r in replacement_str.rstrip(nl).split(nl)]
    len_post = len(replacement_lines)
    if len_pre >= len_post:
        # Replace all lines of pre with lines of post, set any remainder to `None`
        for i in range(len_pre):
            if i in range(len_post):
                post_line = replacement_lines[i]
                lines[pre_startline + i] = post_line
            else:
                lines[pre_startline + i] = None
    else:  # len_pre < len_post
        # Set the original line range to `None`, except for the first line of the
        # range which is replaced with the entire replacement string
        for i in range(len_pre):
            if i > 0:
                lines[pre_startline + i] = None
            else:
                lines[pre_startline] = replacement_str
    return
